(****************** Matematica - corso di LPL: Laboratorio 2020 *****************)
(* ******************************************************************************)
(* Definizione del Linguaggio Small2020: Nucleo di C modificato ed esteso       *)
(* Realizzazione di un Esecutore di Small2020 utilizzando OCaml                 *)
(*                + Laboratorio del 03/04/20                                    *)
(*                + Laboratorio del 07/04/20                                    *)
(*                + Laboratorio del 08/04/20                                    *)
(*                + Laboratorio del 09/04/20                                    *)
(*                + Laboratorio del 19/04/20                                    *)
(*                + Laboratorio del 24/04/20                                    *)
(*                + Da copiare, in Small3, prima di modificare   			    *)
(********************************************************************************)

(* =============================================================================*)
(* Parametri di Programma e Globali *)

let storeSize = 1000;;   (* store size in int/bool words *)


(* ============================================================================ *)
(* The Token, i.e. Lexical Categories *)

type ide = string;;
type num = int;;
type bool = True | False;;



(* ============================================================================ *)
(* The Syntactic Categories *)
(* Vedi Laboratorio2 (slides), laboratorio1: Ocaml del 27/03/20 *)

type tye = 
	Int
	| Bool
	| Void
	| Arr of tye * num
	| Mut of tye
	| Abs of tye * tyeSeq
	| Terr
	and

tyeSeq = tye list
	and	
	
dcl = 
	Const of tye * ide * num
	| Var of tye * ide * exp
	| VarN of tye * ide
	| Array of tye * ide * num
	and 
	
exp = 
	Val of ide
	| N of num
	| B of bool
	| GetArrow of ide * exp
	| Neg of exp
	| Plus of exp * exp
	| Minus of exp * exp
	| Mul of exp * exp
	| Div of exp * exp
	| Eq of exp * exp
	| LT of exp * exp
	| GT of exp * exp
	| Not of exp
	| Or of exp * exp
	| And of exp * exp
	| Upd of exp * exp
	| EE
	and 

cmd = 
	SeqC of cmd * cmd
	| IfE of exp * cmd * cmd
	| IfT of exp * cmd
	| For of stm * exp * stm * cmd
	| Cmd of exp
	| EC
	and 
	
stm = 
	StmD of dcl
	| StmC of cmd
	| Seq of stm list
	;;			

type prog = Prog of ide * stm;;

(* Operations for the AST presentation in the given Concrete Syntax *)
exception Error of string;;

open Printf;; 

let toStringI (i:ide) = sprintf "%s" i;;

let toStringN (n:num) = sprintf "%d" n;;

let toStringB = function
		| True -> "true"
		| False -> "false"
		;;

let rec indent tab = 
		if tab < 0 then ""
		else if tab = 0 then ""
			 else " " ^ (indent (tab-1))
		;;

let rec toStringTye = (function 
		| Int -> "int"
		| Bool -> "bool"
	    | Void -> "void"
	    | Arr(ty,n) -> (toStringTye ty)^ "[" ^(toStringN n)^ "]"
	    | _ -> let msg = "toStringTye: Hide types in AST" in
	    	   raise (Error msg))	
       and

toStringTseq x = (
		let msg = "toStringTye: Hide types in AST" in
	    raise (Error msg)) 
		and 

toStringDcl = (function
		| Const(tye,ide,num) -> ("final " ^ (toStringTye tye) ^ " " 
								 ^ (toStringI ide) ^ " = " ^ (toStringN num))
		| Var(tye,ide,exp) -> ("var " ^ (toStringTye tye) ^ " " 
							   ^ (toStringI ide) ^ " = " ^ (toStringExp exp))
		| VarN(tye,ide) -> ("var " ^ (toStringTye tye) ^ " " ^ (toStringI ide))
		| Array(tye,ide,num) -> ((toStringTye tye) ^ " array " ^ " " 
								 ^ (toStringI ide) ^ "[" ^ (toStringN num) ^ "]"))
		and

toStringExp = (function
		| Val ide -> toStringI ide
	    | N num -> toStringN num
	    | B bool -> toStringB bool
		| GetArrow(ide,exp) -> (toStringI ide) ^ "[" ^ (toStringExp exp) ^ "]"
		| Neg exp -> "-" ^ (toStringExp exp) 
		| Plus(exp1,exp2) -> "("^(toStringExp exp1) ^ " + " ^ (toStringExp exp2)^")"
		| Minus(exp1,exp2) -> "("^(toStringExp exp1) ^ " - " ^ (toStringExp exp2)^")"
		| Mul(exp1,exp2) -> "("^(toStringExp exp1) ^ " * " ^ (toStringExp exp2)^")"
		| Div(exp1,exp2) -> "("^(toStringExp exp1) ^ " / " ^ (toStringExp exp2)^")"
		| Eq(exp1,exp2) -> "("^(toStringExp exp1) ^ " == " ^ (toStringExp exp2)^")"
		| LT(exp1,exp2) -> "("^(toStringExp exp1) ^ " < " ^ (toStringExp exp2)^")"
		| GT(exp1,exp2) -> "("^(toStringExp exp1) ^ " > " ^ (toStringExp exp2)^")"
		| Not exp -> "not " ^ (toStringExp exp)
		| Or(exp1,exp2) -> "("^(toStringExp exp1) ^ " or " ^ (toStringExp exp2)^")"
		| And(exp1,exp2) -> "("^(toStringExp exp1) ^ " and " ^ (toStringExp exp2)^")"
		| Upd(exp1,exp2) -> "("^(toStringExp exp1) ^ " = " ^ (toStringExp exp2)^")"
		| EE -> " ")
		and
		
isAtomic = (function  
		| Val _ -> true
	    | N _ -> true
	    | B _ -> true
		| GetArrow _ -> true
		| Neg _ -> true
		| Not _ -> true
		| _ -> false)
		and

toStringCmd tab = (function
		| SeqC(cmd1,cmd2) -> ((indent tab) ^ "{\n" ^ (flattenC (tab+1) cmd1) 
							  ^ (flattenC (tab+1) cmd2) ^ (indent tab) ^ "}")
		| IfE(exp,cmd1,cmd2) -> (let expString = 
									 (if (isAtomic exp) 
									  then "(" ^ (toStringExp exp) ^ ")" 
									  else toStringExp exp) in
								 (indent tab) ^ "if " ^ expString 
							      ^ (toStringCmd 1 cmd1) ^ ";\n" 
							      ^ (indent tab) ^ "else " ^ (toStringCmd 0 cmd2))
		| IfT(exp,cmd) -> (let expString = 
							   (if (isAtomic exp) 
								then "(" ^ (toStringExp exp) ^ ") " 
								else toStringExp exp) in
						   (indent tab) ^ "if " ^ expString ^ (toStringCmd 1 cmd))
		| For(stm1,exp,stm2,cmd) -> ((indent tab) ^ "for (" ^(toStringStm 0 stm1) ^ ";" 
							     ^ (toStringExp exp) ^ ";" 
							     ^ (toStringStm 0 stm2) ^ ")\n" 
							     ^ (toStringCmd (tab+10) cmd))
		| Cmd exp -> (let s = toStringExp exp in
				      let expString = ( 
					  if isAtomic exp then s
				      else String.sub s 1 ((String.length s)-2)) in
		              (indent tab) ^ expString)
		| EC -> " " )
		and
		
flattenC tab = (function
		| SeqC(cmd1,cmd2) -> (flattenC tab cmd1) ^ (flattenC tab cmd2)
		| EC -> "" 
		| x -> (toStringCmd tab x) ^ ";\n")
		and
		
toStringStm tab = (function
		StmD dcl -> (indent tab) ^ (toStringDcl dcl)
		| StmC cmd -> toStringCmd tab cmd
		| Seq list -> let g = fun s -> (toStringStm tab s)^";\n" in
				  	  let ll = List.map g list in
				      List.fold_left (^) "" ll)
	    and
	
toStringProg (Prog(ide,stm)) = 
		let body = toStringStm 3 stm in
		"Program " ^ (toStringI ide) ^ "{\n" ^ body ^ "}"
				
let printDcl dcl = printf "\n%s\n\n" (toStringDcl dcl);;
let printExp exp = printf "\n%s\n\n" (toStringExp exp);;
let printCmd tab cmd = printf "\n%s\n\n" (toStringCmd tab cmd);;
let printStm tab stm = printf "\n%s\n\n" (toStringStm tab stm);;
let printProg prog = printf "\n%s\n\n" (toStringProg prog);;
		
		
(* ============================================================================ *)
(* The Semantic Structure: Store e Environment *)	
(* Store: Definition *)
(********** Store: Structure *)
type loc = Loc of int;;
type mval = Mint of int 
	| Mbool of bool
	| Undef
	;;
type store = Store of int * (loc -> mval);;
(* Funzione di Astrazione ed Invariante di Rappresentazione per store: 
1) Presentazione (o forma dei valori): [l1<-Mv1,...,lk<-Mvk] per k>=0
2) AF(c) = [] if c=Store(0,g) 
   AF(c) = [l1<-Mv1,...,lk<-Mvk] 
   		   if c=Store(k,g) and (perOgni i\in[1..k], li=Loc(i-1) and (g li)=Mvi).
3) I(Store(n,g)) = 0≤n<=storeSize and 
		(perOgni i\in[1..n], li=Loc(i-1) and (g li) = Mval u and u\in [int]) and
		(perOgni i\in[n+1..storeSize], li=Loc(i-1) and (g li) = Undef
*) 
(********** Store: Operations and Exceptions *)
exception IllegalAddress of string * loc;;
exception UndefinedLoc of string * loc;;
exception StoreSizeExceeded of string * int * int;;

let emptyStore () =  
		Store (0 , fun ((Loc n)as x) -> 
					if (n >= 0) && (n < storeSize) 
					then Undef
					else raise(IllegalAddress("emptyStore",x)))
		;; (* uso di patterns come parametri formali e di "as" per nominarli *)
		
let allocate (Store(d,g)) (k:int) = 
		(Store(d,g)) (* DA CORREGGERE ex1 *)
		;;

let checkLoc = fun (Loc x) d -> x >= 0 && x < d ;; 	(* operazione locale *)	

let upd (Store(d,g)) (loc:loc) (mv:mval) =  
		(Store(d,g)) (* DA CORREGGERE ex2 *)
		;; 

let getStore (Store(d,g)) (loc:loc) =  
		25   (* DA CORREGGERE ex3 *)
		;;

(********** Store: Operation for Store Presentation *)
let toStringMval = function
		Mint n -> sprintf "%d" n 
		| Mbool b -> sprintf "%s" (toStringB b)
		| _ -> "Undef";;

let toStringLoc (Loc n) = sprintf "L%d" n;;

let toStringStore (Store(d,g)) = (* una presentazione dello store *)
		if d = 0 then "[]"
		else let pair n h = 
						(let loc = Loc n in 
						 let v = toStringMval(h loc) and l = toStringLoc loc in
						 l^"<-"^v) in
		     let rec range n m h = (* requires n<m *)
						(if n = m-1 
						then pair n h
						else (pair n h)^","^(range (n+1) m h)) in	
			"["^(range 0 d g)^"]"							
		;;
let storeDump store = 			 (* dump dello store *)
printf "\n%s\n\n" (toStringStore store);;
						
(* Environment: Definition *)
(********** Env: Structure *)
type dval = 
	DConstN of tye * num
	| DConstB of tye * bool
	| DVar of tye * loc
	| DArray of tye * loc
	| Unbound
	;;
type env = Env of ide list * (ide -> dval);;
(* Funzione di Astrazione ed Invariante di Rappresentazione per env.
1) Presentazione (o forma dei valori): [id1/D1,...,idk/Dk] per k>=0
2) AF(c) = [] if c = Env(l,g) and (perOgni id, g(id) = Unbound)
   AF(c) = [id1/D1,...,idk/Dk] 
   		 if c = Env(l,g) and (g(idi) = Di per i\in[1..k], 
   		                      g(x) = Unbound altrimenti).
3) I(c) = (c = Env(l,g) => (List.mem x l) iff (g(x) != Unbound)
*)
(*********** Env: Operations and Exceptions *)
exception UnboundIde of string * ide;;
exception WrongDval of string

open List;;
open String;;

let isSimple = function
	Int -> true
	| Bool -> true
	| _ -> false
	;; 
let isDval = function
	DConstN (Int,(n:num)) -> true
    | DConstB (Bool,(b:bool)) -> true
	| DVar(t,loc) -> isSimple t 
	| DArray(Arr(t,n),loc) -> isSimple t
	| Unbound -> true
	| _ -> false
	;;	
let emptyEnv () = 
		[]   (* DA CORREGGERE ex4 *)
		;;
let bind (Env(l,r)) ide dval =
		Env(l,r)    (* DA CORREGGERE ex5 *)
		;;	
let defined (Env(l,r)) ide =
		mem ide l
		;;		;;
let getEnv (Env(l,r)) ide =
		25     (* DA CORREGGERE ex6 *)
		;;

(********* Env: Operation for env presentation *)

let toStringDval = function
		DConstN (Int,(n:num)) -> sprintf "(int,%d)" n
		| DConstB (Bool,(b:bool)) -> sprintf "(bool,%s)" (toStringB b)
		| DVar(t,loc) when isSimple t 
				-> sprintf "(%s,%s)" (toStringTye t) (toStringLoc loc)
		| DArray(Arr(t,n),loc) when isSimple t
				-> sprintf "(%s,%s)" (toStringTye (Arr(t,n))) 
									 (toStringLoc loc)		  
		| Unbound -> "Unbound"
		| _ -> raise(WrongDval("toStringDval"))
		;;

let toStringEnv (Env (ideList, g)) = (* una presentazione di Env *)
		let pair h i = (let ide = toStringI i and 
		                    den = toStringDval (h i) in ide^"/"^den) in
        let pList = List.map (pair g) ideList in
        let gTemp = (fun x a -> let check = ((get a 0) = ']') in
        			            if check then x^a else (x^"; "^a)) in
        "[" ^ (fold_right gTemp pList "]")
		;;

let envDump env = 			(* dump di env *)
		printf "\n%s\n\n" (toStringEnv env)
		;;

(*********** State: A pair (env,store) -- Operation for state presentation *)
let showState (env,store) =           (* dump dello stato *)
		printf "\nStack:\n%s\nStore:\n%s\n\n" (toStringEnv env) 
											  (toStringStore store) 
		;;

let sigma0 = (emptyEnv(),emptyStore())
		;;



(* ============================================================================ *)
(* The Semantic Function: dclSem, expSem, cmdSem, progSem *)


(* ============================================================================ *)
(* The Semantic Function: dclSem *)


(* ============================================================================ *)
(* The Semantic Function: expSem *)


(* ============================================================================ *)
(* The Semantic Function: cmdSem *)


(* ============================================================================ *)
(* The Semantic Function: progSem *)


(* ============================================================================ *)
(* Tests: Abstract Syntax *)
(* tests in Laboratorio3 *)
(* Program P0 *)
let e1 = Plus(N 3,Val "x");;
let d4 = Var(Int,"z",N 0);;
let e2 = Plus(Val "x",Val "y");;
let c22 = Cmd(Upd(Val "z",Val "y"));;
let c23 = Cmd(Upd(Val "z",N 13));;
let e32 = LT(Val "z",N 12);;
let c341 = Cmd(Upd(GetArrow("A",Val "z"),Plus(Val "y",Val "z")));;
let c342 = Cmd(Upd(Val "z",Plus(Val "z",N 1)));;
let c34 = SeqC(c341,c342);;
let body = Seq [
	StmD(Const(Int,"x",7));
	StmD(Var(Int,"y",e1));
	StmC(Cmd(Upd(Val "y",e2)));
	StmD(VarN(Int,"z"));
	StmC(IfE(LT(e2,N 15),c22,c23));
	StmD(Array(Int,"A",12));
	StmC(For(StmD d4,e32,StmC EC,c34))
	];;
let p0 = Prog("P0",body) in
printProg p0;;

(* ============================================================================ *)
(* Tests: Semantic Structures *)
(******** Tests on Store *)
let mu0 = emptyStore();;
storeDump mu0;;
(*
let (l,mu1) = allocate mu0 3;;
storeDump mu1;;
let mu2 = upd mu0 (Loc 0) (Mbool True);;   (* -- solleva eccezione e ... *
let mu2 = upd (upd mu1 (Loc 2) (Mbool False)) (Loc 0) (Mint (-5));;
storeDump mu2;;
let mu3 = upd mu2 (Loc 1) (Mint 7);;
storeDump mu3;;
toStringMval(getStore mu3 (Loc 0));;
*)
(******** Tests on Env *)
(*
let rho0 = emptyEnv();;
envDump rho0;;
let rho1 = bind rho0 "x" (DArray(Arr(Bool,5),Loc 1));;
envDump rho1;;
let den1 = getEnv rho1 "x";;
toStringDval den1;;
printf "\n%s\n\n" (toStringDval den1);;
let den2 = getEnv rho1 "foo";;   (* -- solleva eccezione e ... *
*)

(* Tests: Semantic Functions *)
(* Tests: Programs and Computations *)
(* Tests: State Dumping *)



